@@ -32,9 +32,9 @@ import (
 
 	. "launchpad.net/gocheck"
 
+	"launchpad.net/ubuntu-push/click"
 	"launchpad.net/ubuntu-push/client/gethosts"
-	"launchpad.net/ubuntu-push/client/session/levelmap"
-	"launchpad.net/ubuntu-push/logger"
+	"launchpad.net/ubuntu-push/client/session/seenstate"
 	"launchpad.net/ubuntu-push/protocol"
 	helpers "launchpad.net/ubuntu-push/testing"
 	"launchpad.net/ubuntu-push/testing/condition"
@@ -157,17 +157,20 @@ func (c *testProtocol) WriteMessage(src interface{}) error {
 	return nil
 }
 
-// brokenLevelMap is a LevelMap that always breaks
-type brokenLevelMap struct{}
+// brokenSeenState is a SeenState that always breaks
+type brokenSeenState struct{}
 
-func (*brokenLevelMap) Set(string, int64) error           { return errors.New("broken.") }
-func (*brokenLevelMap) GetAll() (map[string]int64, error) { return nil, errors.New("broken.") }
+func (*brokenSeenState) SetLevel(string, int64) error            { return errors.New("broken.") }
+func (*brokenSeenState) GetAllLevels() (map[string]int64, error) { return nil, errors.New("broken.") }
+func (*brokenSeenState) FilterBySeen([]protocol.Notification) ([]protocol.Notification, error) {
+	return nil, errors.New("broken.")
+}
 
 /////
 
 type clientSessionSuite struct {
-	log  logger.Logger
-	lvls func() (levelmap.LevelMap, error)
+	log  *helpers.TestLogger
+	lvls func() (seenstate.SeenState, error)
 }
 
 func (cs *clientSessionSuite) SetUpTest(c *C) {
@@ -175,7 +178,7 @@ func (cs *clientSessionSuite) SetUpTest(c *C) {
 }
 
 // in-memory level map testing
-var _ = Suite(&clientSessionSuite{lvls: levelmap.NewLevelMap})
+var _ = Suite(&clientSessionSuite{lvls: seenstate.NewSeenState})
 
 // sqlite level map testing
 type clientSqlevelsSessionSuite struct{ clientSessionSuite }
@@ -183,7 +186,7 @@ type clientSqlevelsSessionSuite struct{ clientSessionSuite }
 var _ = Suite(&clientSqlevelsSessionSuite{})
 
 func (cs *clientSqlevelsSessionSuite) SetUpSuite(c *C) {
-	cs.lvls = func() (levelmap.LevelMap, error) { return levelmap.NewSqliteLevelMap(":memory:") }
+	cs.lvls = func() (seenstate.SeenState, error) { return seenstate.NewSqliteSeenState(":memory:") }
 }
 
 /****************************************************************
@@ -249,8 +252,8 @@ func (cs *clientSessionSuite) TestNewSessionBadPEMFileContentFails(c *C) {
 	c.Check(err, NotNil)
 }
 
-func (cs *clientSessionSuite) TestNewSessionBadLevelMapFails(c *C) {
-	ferr := func() (levelmap.LevelMap, error) { return nil, errors.New("Busted.") }
+func (cs *clientSessionSuite) TestNewSessionBadSeenStateFails(c *C) {
+	ferr := func() (seenstate.SeenState, error) { return nil, errors.New("Busted.") }
 	sess, err := NewSession("", dummyConf, "wah", ferr, cs.log)
 	c.Check(sess, IsNil)
 	c.Assert(err, NotNil)
@@ -346,6 +349,30 @@ func (cs *clientSessionSuite) TestGetHostsRemoteCachingReset(c *C) {
 	c.Check(sess.deliveryHosts, DeepEquals, []string{"baz:443"})
 }
 
+/****************************************************************
+  addAuthorization() tests
+****************************************************************/
+
+func (cs *clientSessionSuite) TestAddAuthorizationAddsAuthorization(c *C) {
+	url := "xyzzy://"
+	sess := &ClientSession{Log: cs.log}
+	sess.AuthGetter = func(url string) string { return url + " auth'ed" }
+	sess.AuthURL = url
+	c.Assert(sess.auth, Equals, "")
+	err := sess.addAuthorization()
+	c.Assert(err, IsNil)
+	c.Check(sess.auth, Equals, "xyzzy:// auth'ed")
+}
+
+func (cs *clientSessionSuite) TestAddAuthorizationSkipsIfUnset(c *C) {
+	sess := &ClientSession{Log: cs.log}
+	sess.AuthGetter = nil
+	c.Assert(sess.auth, Equals, "")
+	err := sess.addAuthorization()
+	c.Assert(err, IsNil)
+	c.Check(sess.auth, Equals, "")
+}
+
 /****************************************************************
   startConnectionAttempt()/nextHostToTry()/started tests
 ****************************************************************/
@@ -601,7 +628,7 @@ func (s *msgSuite) SetUpTest(c *C) {
 	conf := ClientSessionConfig{
 		ExchangeTimeout: time.Millisecond,
 	}
-	s.sess, err = NewSession("", conf, "wah", levelmap.NewLevelMap, helpers.NewTestLogger(c, "debug"))
+	s.sess, err = NewSession("", conf, "wah", seenstate.NewSeenState, helpers.NewTestLogger(c, "debug"))
 	c.Assert(err, IsNil)
 	s.sess.Connection = &testConn{Name: "TestHandle*"}
 	s.errCh = make(chan error, 1)
@@ -609,7 +636,8 @@ func (s *msgSuite) SetUpTest(c *C) {
 	s.downCh = make(chan interface{}, 5)
 	s.sess.proto = &testProtocol{up: s.upCh, down: s.downCh}
 	// make the message channel buffered
-	s.sess.MsgCh = make(chan *Notification, 5)
+	s.sess.BroadcastCh = make(chan *BroadcastNotification, 5)
+	s.sess.NotificationsCh = make(chan AddressedNotification, 5)
 }
 
 func (s *msgSuite) TestHandlePingWorks(c *C) {
@@ -652,24 +680,25 @@ func (s *msgSuite) TestHandlePingDoesNotClearsDelayOnError(c *C) {
 ****************************************************************/
 
 func (s *msgSuite) TestHandleBroadcastWorks(c *C) {
-	msg := serverMsg{"broadcast",
-		protocol.BroadcastMsg{
-			Type:     "broadcast",
-			AppId:    "--ignored--",
-			ChanId:   "0",
-			TopLevel: 2,
-			Payloads: []json.RawMessage{
-				json.RawMessage(`{"img1/m1":[101,"tubular"]}`),
-				json.RawMessage("false"), // shouldn't happen but robust
-				json.RawMessage(`{"img1/m1":[102,"tubular"]}`),
-			},
-		}, protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+	msg := new(serverMsg)
+	msg.Type = "broadcast"
+	msg.BroadcastMsg = protocol.BroadcastMsg{
+		Type:     "broadcast",
+		AppId:    "--ignored--",
+		ChanId:   "0",
+		TopLevel: 2,
+		Payloads: []json.RawMessage{
+			json.RawMessage(`{"img1/m1":[101,"tubular"]}`),
+			json.RawMessage("false"), // shouldn't happen but robust
+			json.RawMessage(`{"img1/m1":[102,"tubular"]}`),
+		},
+	}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
 	s.upCh <- nil // ack ok
 	c.Check(<-s.errCh, Equals, nil)
-	c.Assert(len(s.sess.MsgCh), Equals, 1)
-	c.Check(<-s.sess.MsgCh, DeepEquals, &Notification{
+	c.Assert(len(s.sess.BroadcastCh), Equals, 1)
+	c.Check(<-s.sess.BroadcastCh, DeepEquals, &BroadcastNotification{
 		TopLevel: 2,
 		Decoded: []map[string]interface{}{
 			map[string]interface{}{
@@ -681,21 +710,22 @@ func (s *msgSuite) TestHandleBroadcastWorks(c *C) {
 		},
 	})
 	// and finally, the session keeps track of the levels
-	levels, err := s.sess.Levels.GetAll()
+	levels, err := s.sess.SeenState.GetAllLevels()
 	c.Check(err, IsNil)
 	c.Check(levels, DeepEquals, map[string]int64{"0": 2})
 }
 
 func (s *msgSuite) TestHandleBroadcastBadAckWrite(c *C) {
-	msg := serverMsg{"broadcast",
-		protocol.BroadcastMsg{
-			Type:     "broadcast",
-			AppId:    "APP",
-			ChanId:   "0",
-			TopLevel: 2,
-			Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
-		}, protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+	msg := new(serverMsg)
+	msg.Type = "broadcast"
+	msg.BroadcastMsg = protocol.BroadcastMsg{
+		Type:     "broadcast",
+		AppId:    "APP",
+		ChanId:   "0",
+		TopLevel: 2,
+		Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
+	}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
 	failure := errors.New("ACK ACK ACK")
 	s.upCh <- failure
@@ -704,37 +734,40 @@ func (s *msgSuite) TestHandleBroadcastBadAckWrite(c *C) {
 }
 
 func (s *msgSuite) TestHandleBroadcastWrongChannel(c *C) {
-	msg := serverMsg{"broadcast",
-		protocol.BroadcastMsg{
-			Type:     "broadcast",
-			AppId:    "APP",
-			ChanId:   "something awful",
-			TopLevel: 2,
-			Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
-		}, protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+	msg := new(serverMsg)
+	msg.Type = "brodacast"
+	msg.BroadcastMsg = protocol.BroadcastMsg{
+		Type:     "broadcast",
+		AppId:    "APP",
+		ChanId:   "something awful",
+		TopLevel: 2,
+		Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
+	}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
 	s.upCh <- nil // ack ok
 	c.Check(<-s.errCh, IsNil)
-	c.Check(len(s.sess.MsgCh), Equals, 0)
+	c.Check(len(s.sess.BroadcastCh), Equals, 0)
 }
 
-func (s *msgSuite) TestHandleBroadcastWrongBrokenLevelmap(c *C) {
-	s.sess.Levels = &brokenLevelMap{}
-	msg := serverMsg{"broadcast",
-		protocol.BroadcastMsg{
-			Type:     "broadcast",
-			AppId:    "--ignored--",
-			ChanId:   "0",
-			TopLevel: 2,
-			Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
-		}, protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+func (s *msgSuite) TestHandleBroadcastBrokenSeenState(c *C) {
+	s.sess.SeenState = &brokenSeenState{}
+	msg := new(serverMsg)
+	msg.Type = "broadcast"
+	msg.BroadcastMsg = protocol.BroadcastMsg{
+		Type:     "broadcast",
+		AppId:    "--ignored--",
+		ChanId:   "0",
+		TopLevel: 2,
+		Payloads: []json.RawMessage{json.RawMessage(`{"b":1}`)},
+	}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	s.upCh <- nil // ack ok
 	// start returns with error
 	c.Check(<-s.errCh, Not(Equals), nil)
+	c.Check(s.sess.State(), Equals, Error)
 	// no message sent out
-	c.Check(len(s.sess.MsgCh), Equals, 0)
+	c.Check(len(s.sess.BroadcastCh), Equals, 0)
 	// and nak'ed it
 	c.Check(len(s.downCh), Equals, 1)
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"nak"})
@@ -743,9 +776,8 @@ func (s *msgSuite) TestHandleBroadcastWrongBrokenLevelmap(c *C) {
 func (s *msgSuite) TestHandleBroadcastClearsDelay(c *C) {
 	s.sess.setShouldDelay()
 
-	msg := serverMsg{"broadcast", protocol.BroadcastMsg{},
-		protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+	msg := &serverMsg{Type: "broadcast"}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
 	s.upCh <- nil // ack ok
 	c.Check(<-s.errCh, IsNil)
@@ -756,9 +788,8 @@ func (s *msgSuite) TestHandleBroadcastClearsDelay(c *C) {
 func (s *msgSuite) TestHandleBroadcastDoesNotClearDelayOnError(c *C) {
 	s.sess.setShouldDelay()
 
-	msg := serverMsg{"broadcast", protocol.BroadcastMsg{},
-		protocol.NotificationsMsg{}, protocol.ConnBrokenMsg{}}
-	go func() { s.errCh <- s.sess.handleBroadcast(&msg) }()
+	msg := &serverMsg{Type: "broadcast"}
+	go func() { s.errCh <- s.sess.handleBroadcast(msg) }()
 	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
 	s.upCh <- errors.New("bcast")
 	c.Check(<-s.errCh, NotNil)
@@ -766,31 +797,232 @@ func (s *msgSuite) TestHandleBroadcastDoesNotClearDelayOnError(c *C) {
 	c.Check(s.sess.ShouldDelay(), Equals, true)
 }
 
+/****************************************************************
+  handleNotifications() tests
+****************************************************************/
+
+type testAddresseeChecking struct {
+	ops     chan string
+	missing string
+}
+
+func (ac *testAddresseeChecking) StartAddresseeBatch() {
+	ac.ops <- "start"
+}
+
+func (ac *testAddresseeChecking) CheckForAddressee(notif *protocol.Notification) *click.AppId {
+	ac.ops <- notif.AppId
+	if notif.AppId != ac.missing {
+		id, err := click.ParseAppId(notif.AppId)
+		if err != nil {
+			panic(err)
+		}
+		return id
+	} else {
+		return nil
+	}
+}
+
+func (s *msgSuite) TestHandleNotificationsWorks(c *C) {
+	ac := &testAddresseeChecking{ops: make(chan string, 10)}
+	s.sess.AddresseeChecker = ac
+	s.sess.setShouldDelay()
+	n1 := protocol.Notification{
+		AppId:   "com.example.app1_app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	n2 := protocol.Notification{
+		AppId:   "com.example.app2_app2",
+		MsgId:   "b",
+		Payload: json.RawMessage(`{"m": 2}`),
+	}
+	msg := new(serverMsg)
+	msg.Type = "notifications"
+	msg.NotificationsMsg = protocol.NotificationsMsg{
+		Notifications: []protocol.Notification{n1, n2},
+	}
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	s.upCh <- nil // ack ok
+	c.Check(<-s.errCh, Equals, nil)
+	c.Check(s.sess.ShouldDelay(), Equals, false)
+	c.Assert(s.sess.NotificationsCh, HasLen, 2)
+	app1, err := click.ParseAppId("com.example.app1_app1")
+	c.Assert(err, IsNil)
+	c.Check(<-s.sess.NotificationsCh, DeepEquals, AddressedNotification{
+		To:           app1,
+		Notification: &n1,
+	})
+	app2, err := click.ParseAppId("com.example.app2_app2")
+	c.Assert(err, IsNil)
+	c.Check(<-s.sess.NotificationsCh, DeepEquals, AddressedNotification{
+		To:           app2,
+		Notification: &n2,
+	})
+	c.Check(ac.ops, HasLen, 3)
+	c.Check(<-ac.ops, Equals, "start")
+	c.Check(<-ac.ops, Equals, "com.example.app1_app1")
+	c.Check(<-ac.ops, Equals, "com.example.app2_app2")
+}
+
+func (s *msgSuite) TestHandleNotificationsAddresseeCheck(c *C) {
+	ac := &testAddresseeChecking{
+		ops:     make(chan string, 10),
+		missing: "com.example.app1_app1",
+	}
+	s.sess.AddresseeChecker = ac
+	s.sess.setShouldDelay()
+	n1 := protocol.Notification{
+		AppId:   "com.example.app1_app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	n2 := protocol.Notification{
+		AppId:   "com.example.app2_app2",
+		MsgId:   "b",
+		Payload: json.RawMessage(`{"m": 2}`),
+	}
+	msg := new(serverMsg)
+	msg.Type = "notifications"
+	msg.NotificationsMsg = protocol.NotificationsMsg{
+		Notifications: []protocol.Notification{n1, n2},
+	}
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	s.upCh <- nil // ack ok
+	c.Check(<-s.errCh, Equals, nil)
+	c.Check(s.sess.ShouldDelay(), Equals, false)
+	c.Assert(s.sess.NotificationsCh, HasLen, 1)
+	app2, err := click.ParseAppId("com.example.app2_app2")
+	c.Assert(err, IsNil)
+	c.Check(<-s.sess.NotificationsCh, DeepEquals, AddressedNotification{
+		To:           app2,
+		Notification: &n2,
+	})
+	c.Check(ac.ops, HasLen, 3)
+	c.Check(<-ac.ops, Equals, "start")
+	c.Check(<-ac.ops, Equals, "com.example.app1_app1")
+}
+
+func (s *msgSuite) TestHandleNotificationsFiltersSeen(c *C) {
+	ac := &testAddresseeChecking{ops: make(chan string, 10)}
+	s.sess.AddresseeChecker = ac
+	n1 := protocol.Notification{
+		AppId:   "com.example.app1_app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	n2 := protocol.Notification{
+		AppId:   "com.example.app2_app2",
+		MsgId:   "b",
+		Payload: json.RawMessage(`{"m": 2}`),
+	}
+	msg := new(serverMsg)
+	msg.Type = "notifications"
+	msg.NotificationsMsg = protocol.NotificationsMsg{
+		Notifications: []protocol.Notification{n1, n2},
+	}
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	s.upCh <- nil // ack ok
+	c.Check(<-s.errCh, Equals, nil)
+	c.Assert(s.sess.NotificationsCh, HasLen, 2)
+	app1, err := click.ParseAppId("com.example.app1_app1")
+	c.Assert(err, IsNil)
+	c.Check(<-s.sess.NotificationsCh, DeepEquals, AddressedNotification{
+		To:           app1,
+		Notification: &n1,
+	})
+	app2, err := click.ParseAppId("com.example.app2_app2")
+	c.Assert(err, IsNil)
+	c.Check(<-s.sess.NotificationsCh, DeepEquals, AddressedNotification{
+		To:           app2,
+		Notification: &n2,
+	})
+	c.Check(ac.ops, HasLen, 3)
+
+	// second time they get ignored
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	s.upCh <- nil // ack ok
+	c.Check(<-s.errCh, Equals, nil)
+	c.Assert(s.sess.NotificationsCh, HasLen, 0)
+	c.Check(ac.ops, HasLen, 4)
+}
+
+func (s *msgSuite) TestHandleNotificationsBadAckWrite(c *C) {
+	s.sess.setShouldDelay()
+	n1 := protocol.Notification{
+		AppId:   "com.example.app1_app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	msg := new(serverMsg)
+	msg.Type = "notifications"
+	msg.NotificationsMsg = protocol.NotificationsMsg{
+		Notifications: []protocol.Notification{n1},
+	}
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	failure := errors.New("ACK ACK ACK")
+	s.upCh <- failure
+	c.Assert(<-s.errCh, Equals, failure)
+	c.Check(s.sess.State(), Equals, Error)
+	// didn't get to clear
+	c.Check(s.sess.ShouldDelay(), Equals, true)
+}
+
+func (s *msgSuite) TestHandleNotificationsBrokenSeenState(c *C) {
+	s.sess.setShouldDelay()
+	s.sess.SeenState = &brokenSeenState{}
+	n1 := protocol.Notification{
+		AppId:   "com.example.app1_app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	msg := new(serverMsg)
+	msg.Type = "notifications"
+	msg.NotificationsMsg = protocol.NotificationsMsg{
+		Notifications: []protocol.Notification{n1},
+	}
+	go func() { s.errCh <- s.sess.handleNotifications(msg) }()
+	s.upCh <- nil // ack ok
+	// start returns with error
+	c.Check(<-s.errCh, Not(Equals), nil)
+	c.Check(s.sess.State(), Equals, Error)
+	// no message sent out
+	c.Check(len(s.sess.NotificationsCh), Equals, 0)
+	// and nak'ed it
+	c.Check(len(s.downCh), Equals, 1)
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"nak"})
+	// didn't get to clear
+	c.Check(s.sess.ShouldDelay(), Equals, true)
+}
+
 /****************************************************************
   handleConnBroken() tests
 ****************************************************************/
 
 func (s *msgSuite) TestHandleConnBrokenUnkwown(c *C) {
-	msg := serverMsg{"connbroken",
-		protocol.BroadcastMsg{}, protocol.NotificationsMsg{},
-		protocol.ConnBrokenMsg{
-			Reason: "REASON",
-		},
+	msg := new(serverMsg)
+	msg.Type = "connbroken"
+	msg.ConnBrokenMsg = protocol.ConnBrokenMsg{
+		Reason: "REASON",
 	}
-	go func() { s.errCh <- s.sess.handleConnBroken(&msg) }()
+	go func() { s.errCh <- s.sess.handleConnBroken(msg) }()
 	c.Check(<-s.errCh, ErrorMatches, "server broke connection: REASON")
 	c.Check(s.sess.State(), Equals, Error)
 }
 
 func (s *msgSuite) TestHandleConnBrokenHostMismatch(c *C) {
-	msg := serverMsg{"connbroken",
-		protocol.BroadcastMsg{}, protocol.NotificationsMsg{},
-		protocol.ConnBrokenMsg{
-			Reason: protocol.BrokenHostMismatch,
-		},
+	msg := new(serverMsg)
+	msg.Type = "connbroken"
+	msg.ConnBrokenMsg = protocol.ConnBrokenMsg{
+		Reason: protocol.BrokenHostMismatch,
 	}
 	s.sess.deliveryHosts = []string{"foo:443", "bar:443"}
-	go func() { s.errCh <- s.sess.handleConnBroken(&msg) }()
+	go func() { s.errCh <- s.sess.handleConnBroken(msg) }()
 	c.Check(<-s.errCh, ErrorMatches, "server broke connection: host-mismatch")
 	c.Check(s.sess.State(), Equals, Error)
 	// hosts were reset
@@ -861,6 +1093,40 @@ func (s *loopSuite) TestLoopBroadcast(c *C) {
 	c.Check(<-s.errCh, Equals, failure)
 }
 
+func (s *loopSuite) TestLoopNotifications(c *C) {
+	c.Check(s.sess.State(), Equals, Running)
+
+	n1 := protocol.Notification{
+		AppId:   "app1",
+		MsgId:   "a",
+		Payload: json.RawMessage(`{"m": 1}`),
+	}
+	msg := &protocol.NotificationsMsg{
+		Type:          "notifications",
+		Notifications: []protocol.Notification{n1},
+	}
+	c.Check(takeNext(s.downCh), Equals, "deadline 1ms")
+	s.upCh <- msg
+	c.Check(takeNext(s.downCh), Equals, protocol.AckMsg{"ack"})
+	failure := errors.New("ack")
+	s.upCh <- failure
+	c.Check(<-s.errCh, Equals, failure)
+}
+
+func (s *loopSuite) TestLoopSetParams(c *C) {
+	c.Check(s.sess.State(), Equals, Running)
+	setParams := protocol.SetParamsMsg{
+		Type:      "setparams",
+		SetCookie: "COOKIE",
+	}
+	c.Check(takeNext(s.downCh), Equals, "deadline 1ms")
+	s.upCh <- setParams
+	failure := errors.New("fail")
+	s.upCh <- failure
+	c.Assert(<-s.errCh, Equals, failure)
+	c.Check(s.sess.getCookie(), Equals, "COOKIE")
+}
+
 func (s *loopSuite) TestLoopConnBroken(c *C) {
 	c.Check(s.sess.State(), Equals, Running)
 	broken := protocol.ConnBrokenMsg{
@@ -872,6 +1138,31 @@ func (s *loopSuite) TestLoopConnBroken(c *C) {
 	c.Check(<-s.errCh, NotNil)
 }
 
+func (s *loopSuite) TestLoopConnWarn(c *C) {
+	warn := protocol.ConnWarnMsg{
+		Type:   "warn",
+		Reason: "XXX",
+	}
+	connwarn := protocol.ConnWarnMsg{
+		Type:   "connwarn",
+		Reason: "REASON",
+	}
+	failure := errors.New("warn")
+	log := s.sess.Log.(*helpers.TestLogger)
+
+	c.Check(s.sess.State(), Equals, Running)
+	c.Check(takeNext(s.downCh), Equals, "deadline 1ms")
+	log.ResetCapture()
+	s.upCh <- warn
+	s.upCh <- connwarn
+	s.upCh <- failure
+	c.Check(<-s.errCh, Equals, failure)
+	c.Check(log.Captured(),
+		Matches, `(?ms).* warning: XXX$.*`)
+	c.Check(log.Captured(),
+		Matches, `(?ms).* warning: REASON$`)
+}
+
 /****************************************************************
   start() tests
 ****************************************************************/
@@ -898,7 +1189,7 @@ func (cs *clientSessionSuite) TestStartFailsIfWriteFails(c *C) {
 func (cs *clientSessionSuite) TestStartFailsIfGetLevelsFails(c *C) {
 	sess, err := NewSession("", dummyConf, "wah", cs.lvls, cs.log)
 	c.Assert(err, IsNil)
-	sess.Levels = &brokenLevelMap{}
+	sess.SeenState = &brokenSeenState{}
 	sess.Connection = &testConn{Name: "TestStartConnectMessageFails"}
 	errCh := make(chan error, 1)
 	upCh := make(chan interface{}, 5)
@@ -931,9 +1222,10 @@ func (cs *clientSessionSuite) TestStartConnectMessageFails(c *C) {
 
 	c.Check(takeNext(downCh), Equals, "deadline 0")
 	c.Check(takeNext(downCh), DeepEquals, protocol.ConnectMsg{
-		Type:     "connect",
-		DeviceId: sess.DeviceId,
-		Levels:   map[string]int64{},
+		Type:          "connect",
+		DeviceId:      sess.DeviceId,
+		Levels:        map[string]int64{},
+		Authorization: "",
 	})
 	upCh <- errors.New("Overflow error in /dev/null")
 	err = <-errCh
@@ -1024,6 +1316,7 @@ func (cs *clientSessionSuite) TestStartWorks(c *C) {
 	sess, err := NewSession("", conf, "wah", cs.lvls, cs.log)
 	c.Assert(err, IsNil)
 	sess.Connection = &testConn{Name: "TestStartWorks"}
+	sess.setCookie("COOKIE")
 	errCh := make(chan error, 1)
 	upCh := make(chan interface{}, 5)
 	downCh := make(chan interface{}, 5)
@@ -1038,6 +1331,8 @@ func (cs *clientSessionSuite) TestStartWorks(c *C) {
 	msg, ok := takeNext(downCh).(protocol.ConnectMsg)
 	c.Check(ok, Equals, true)
 	c.Check(msg.DeviceId, Equals, "wah")
+	c.Check(msg.Authorization, Equals, "")
+	c.Check(msg.Cookie, Equals, "COOKIE")
 	c.Check(msg.Info, DeepEquals, info)
 	upCh <- nil // no error
 	upCh <- protocol.ConnAckMsg{
@@ -1054,6 +1349,22 @@ func (cs *clientSessionSuite) TestStartWorks(c *C) {
   run() tests
 ****************************************************************/
 
+func (cs *clientSessionSuite) TestRunBailsIfAuthCheckFails(c *C) {
+	sess, err := NewSession("", dummyConf, "wah", cs.lvls, cs.log)
+	c.Assert(err, IsNil)
+	failure := errors.New("TestRunBailsIfAuthCheckFails")
+	has_closed := false
+	err = sess.run(
+		func() { has_closed = true },
+		func() error { return failure },
+		nil,
+		nil,
+		nil,
+		nil)
+	c.Check(err, Equals, failure)
+	c.Check(has_closed, Equals, true)
+}
+
 func (cs *clientSessionSuite) TestRunBailsIfHostGetterFails(c *C) {
 	sess, err := NewSession("", dummyConf, "wah", cs.lvls, cs.log)
 	c.Assert(err, IsNil)
@@ -1061,6 +1372,7 @@ func (cs *clientSessionSuite) TestRunBailsIfHostGetterFails(c *C) {
 	has_closed := false
 	err = sess.run(
 		func() { has_closed = true },
+		func() error { return nil },
 		func() error { return failure },
 		nil,
 		nil,
@@ -1076,6 +1388,7 @@ func (cs *clientSessionSuite) TestRunBailsIfConnectFails(c *C) {
 	err = sess.run(
 		func() {},
 		func() error { return nil },
+		func() error { return nil },
 		func() error { return failure },
 		nil,
 		nil)
@@ -1090,6 +1403,7 @@ func (cs *clientSessionSuite) TestRunBailsIfStartFails(c *C) {
 		func() {},
 		func() error { return nil },
 		func() error { return nil },
+		func() error { return nil },
 		func() error { return failure },
 		nil)
 	c.Check(err, Equals, failure)
@@ -1098,23 +1412,24 @@ func (cs *clientSessionSuite) TestRunBailsIfStartFails(c *C) {
 func (cs *clientSessionSuite) TestRunRunsEvenIfLoopFails(c *C) {
 	sess, err := NewSession("", dummyConf, "wah", cs.lvls, cs.log)
 	c.Assert(err, IsNil)
-	// just to make a point: until here we haven't set ErrCh & MsgCh (no
+	// just to make a point: until here we haven't set ErrCh & BroadcastCh (no
 	// biggie if this stops being true)
 	c.Check(sess.ErrCh, IsNil)
-	c.Check(sess.MsgCh, IsNil)
+	c.Check(sess.BroadcastCh, IsNil)
 	failureCh := make(chan error) // must be unbuffered
-	notf := &Notification{}
+	notf := &BroadcastNotification{}
 	err = sess.run(
 		func() {},
 		func() error { return nil },
 		func() error { return nil },
 		func() error { return nil },
-		func() error { sess.MsgCh <- notf; return <-failureCh })
+		func() error { return nil },
+		func() error { sess.BroadcastCh <- notf; return <-failureCh })
 	c.Check(err, Equals, nil)
 	// if run doesn't error it sets up the channels
 	c.Assert(sess.ErrCh, NotNil)
-	c.Assert(sess.MsgCh, NotNil)
-	c.Check(<-sess.MsgCh, Equals, notf)
+	c.Assert(sess.BroadcastCh, NotNil)
+	c.Check(<-sess.BroadcastCh, Equals, notf)
 	failure := errors.New("TestRunRunsEvenIfLoopFails")
 	failureCh <- failure
 	c.Check(<-sess.ErrCh, Equals, failure)
@@ -1174,14 +1489,7 @@ var (
 
 func (cs *clientSessionSuite) TestDialBadServerName(c *C) {
 	// a borked server name
-	cert, err := tls.X509KeyPair(helpers.TestCertPEMBlock, helpers.TestKeyPEMBlock)
-	c.Assert(err, IsNil)
-	tlsCfg := &tls.Config{
-		Certificates:           []tls.Certificate{cert},
-		SessionTicketsDisabled: true,
-	}
-
-	lst, err := tls.Listen("tcp", "localhost:0", tlsCfg)
+	lst, err := tls.Listen("tcp", "localhost:0", helpers.TestTLSServerConfig)
 	c.Assert(err, IsNil)
 	// advertise
 	ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
@@ -1226,19 +1534,12 @@ func (cs *clientSessionSuite) TestDialBadServerName(c *C) {
 
 func (cs *clientSessionSuite) TestDialWorks(c *C) {
 	// happy path thoughts
-	cert, err := tls.X509KeyPair(helpers.TestCertPEMBlock, helpers.TestKeyPEMBlock)
-	c.Assert(err, IsNil)
-	tlsCfg := &tls.Config{
-		Certificates:           []tls.Certificate{cert},
-		SessionTicketsDisabled: true,
-	}
-
-	lst, err := tls.Listen("tcp", "localhost:0", tlsCfg)
+	lst, err := tls.Listen("tcp", "localhost:0", helpers.TestTLSServerConfig)
 	c.Assert(err, IsNil)
 	// advertise
 	ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		b, err := json.Marshal(map[string]interface{}{
-			"domain": "localhost",
+			"domain": "push-delivery",
 			"hosts":  []string{"nowhere", lst.Addr().String()},
 		})
 		if err != nil {
@@ -1317,9 +1618,9 @@ func (cs *clientSessionSuite) TestDialWorks(c *C) {
 	c.Check(takeNext(downCh), Equals, protocol.AckMsg{"ack"})
 	upCh <- nil
 	// ...get bubbled up,
-	c.Check(<-sess.MsgCh, NotNil)
+	c.Check(<-sess.BroadcastCh, NotNil)
 	// and their TopLevel remembered
-	levels, err := sess.Levels.GetAll()
+	levels, err := sess.SeenState.GetAllLevels()
 	c.Check(err, IsNil)
 	c.Check(levels, DeepEquals, map[string]int64{"0": 2})
 
@@ -1334,14 +1635,7 @@ func (cs *clientSessionSuite) TestDialWorks(c *C) {
 
 func (cs *clientSessionSuite) TestDialWorksDirect(c *C) {
 	// happy path thoughts
-	cert, err := tls.X509KeyPair(helpers.TestCertPEMBlock, helpers.TestKeyPEMBlock)
-	c.Assert(err, IsNil)
-	tlsCfg := &tls.Config{
-		Certificates:           []tls.Certificate{cert},
-		SessionTicketsDisabled: true,
-	}
-
-	lst, err := tls.Listen("tcp", "localhost:0", tlsCfg)
+	lst, err := tls.Listen("tcp", "localhost:0", helpers.TestTLSServerConfig)
 	c.Assert(err, IsNil)
 	sess, err := NewSession(lst.Addr().String(), dialTestConf, "wah", cs.lvls, cs.log)
 	c.Assert(err, IsNil)
