@@ -130,8 +130,8 @@ func newTestBroker() *testBroker {
 	return &testBroker{registration: make(chan interface{}, 2)}
 }
 
-func (tb *testBroker) Register(connect *protocol.ConnectMsg) (broker.BrokerSession, error) {
-	tb.registration <- "register " + connect.DeviceId
+func (tb *testBroker) Register(connect *protocol.ConnectMsg, sessionId string) (broker.BrokerSession, error) {
+	tb.registration <- fmt.Sprintf("register %s %s", connect.DeviceId, sessionId)
 	return &testing.TestBrokerSession{DeviceId: connect.DeviceId}, tb.err
 }
 
@@ -148,7 +148,7 @@ func (s *sessionSuite) TestSessionStart(c *C) {
 	brkr := newTestBroker()
 	go func() {
 		var err error
-		sess, err = sessionStart(tp, brkr, cfg10msPingInterval5msExchangeTout)
+		sess, err = sessionStart(tp, brkr, cfg10msPingInterval5msExchangeTout, "s1")
 		errCh <- err
 	}()
 	c.Check(takeNext(down), Equals, "deadline 5ms")
@@ -160,7 +160,7 @@ func (s *sessionSuite) TestSessionStart(c *C) {
 	up <- nil // no write error
 	err := <-errCh
 	c.Check(err, IsNil)
-	c.Check(takeNext(brkr.registration), Equals, "register dev-1")
+	c.Check(takeNext(brkr.registration), Equals, "register dev-1 s1")
 	c.Check(sess.DeviceIdentifier(), Equals, "dev-1")
 }
 
@@ -175,7 +175,7 @@ func (s *sessionSuite) TestSessionRegisterError(c *C) {
 	brkr.err = errRegister
 	go func() {
 		var err error
-		sess, err = sessionStart(tp, brkr, cfg10msPingInterval5msExchangeTout)
+		sess, err = sessionStart(tp, brkr, cfg10msPingInterval5msExchangeTout, "s2")
 		errCh <- err
 	}()
 	up <- protocol.ConnectMsg{Type: "connect", ClientVer: "1", DeviceId: "dev-1"}
@@ -190,7 +190,7 @@ func (s *sessionSuite) TestSessionStartReadError(c *C) {
 	down := make(chan interface{}, 5)
 	tp := &testProtocol{up, down}
 	up <- io.ErrUnexpectedEOF
-	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout)
+	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout, "s3")
 	c.Check(err, Equals, io.ErrUnexpectedEOF)
 }
 
@@ -200,7 +200,7 @@ func (s *sessionSuite) TestSessionStartWriteError(c *C) {
 	tp := &testProtocol{up, down}
 	up <- protocol.ConnectMsg{Type: "connect"}
 	up <- io.ErrUnexpectedEOF
-	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout)
+	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout, "s4")
 	c.Check(err, Equals, io.ErrUnexpectedEOF)
 	// sanity
 	c.Check(takeNext(down), Matches, "deadline.*")
@@ -212,35 +212,38 @@ func (s *sessionSuite) TestSessionStartMismatch(c *C) {
 	down := make(chan interface{}, 5)
 	tp := &testProtocol{up, down}
 	up <- protocol.ConnectMsg{Type: "what"}
-	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout)
+	_, err := sessionStart(tp, nil, cfg10msPingInterval5msExchangeTout, "s5")
 	c.Check(err, DeepEquals, &broker.ErrAbort{"expected CONNECT message"})
 }
 
-var cfg5msPingInterval2msExchangeTout = &testSessionConfig{
-	pingInterval:    5 * time.Millisecond,
-	exchangeTimeout: 2 * time.Millisecond,
-}
-
 func (s *sessionSuite) TestSessionLoop(c *C) {
-	nopTrack := NewTracker(s.testlog)
+	track := &testTracker{NewTracker(s.testlog), make(chan interface{}, 2)}
 	errCh := make(chan error, 1)
 	up := make(chan interface{}, 5)
 	down := make(chan interface{}, 5)
 	tp := &testProtocol{up, down}
 	sess := &testing.TestBrokerSession{}
 	go func() {
-		errCh <- sessionLoop(tp, sess, cfg5msPingInterval2msExchangeTout, nopTrack)
+		errCh <- sessionLoop(tp, sess, cfg10msPingInterval5msExchangeTout, track)
 	}()
-	c.Check(takeNext(down), Equals, "deadline 2ms")
+	c.Check(takeNext(down), Equals, "deadline 5ms")
 	c.Check(takeNext(down), DeepEquals, protocol.PingPongMsg{Type: "ping"})
 	up <- nil // no write error
 	up <- protocol.PingPongMsg{Type: "pong"}
-	c.Check(takeNext(down), Equals, "deadline 2ms")
+	c.Check(takeNext(down), Equals, "deadline 5ms")
 	c.Check(takeNext(down), DeepEquals, protocol.PingPongMsg{Type: "ping"})
 	up <- nil // no write error
 	up <- io.ErrUnexpectedEOF
 	err := <-errCh
 	c.Check(err, Equals, io.ErrUnexpectedEOF)
+	c.Check(track.interval, HasLen, 2)
+	c.Check((<-track.interval).(time.Duration) <= 16*time.Millisecond, Equals, true)
+	c.Check((<-track.interval).(time.Duration) <= 16*time.Millisecond, Equals, true)
+}
+
+var cfg5msPingInterval2msExchangeTout = &testSessionConfig{
+	pingInterval:    5 * time.Millisecond,
+	exchangeTimeout: 2 * time.Millisecond,
 }
 
 func (s *sessionSuite) TestSessionLoopWriteError(c *C) {
@@ -357,11 +360,63 @@ func (s *sessionSuite) TestSessionLoopKick(c *C) {
 	go func() {
 		errCh <- sessionLoop(tp, sess, cfg5msPingInterval2msExchangeTout, nopTrack)
 	}()
-	close(exchanges)
+	exchanges <- nil
 	err := <-errCh
 	c.Check(err, DeepEquals, &broker.ErrAbort{"terminated"})
 }
 
+func (s *sessionSuite) TestPingTimerReset(c *C) {
+	pingInterval := 50 * time.Millisecond
+	now := time.Now()
+	l := &loop{
+		pingInterval:  pingInterval,
+		pingTimer:     time.NewTimer(pingInterval),
+		intervalStart: now,
+	}
+	time.Sleep(10 * time.Millisecond)
+	l.pingTimer.Stop()
+	done := l.pingTimerReset(true)
+	c.Assert(done, Equals, true)
+	select {
+	case <-l.pingTimer.C:
+	case <-time.After(1 * time.Second):
+		c.Fatal("timer should have triggered")
+	}
+	c.Check(now, Not(Equals), l.intervalStart)
+}
+
+func (s *sessionSuite) TestPingTimerResetPartial(c *C) {
+	pingInterval := 5 * time.Second
+	now := time.Now().Add(-pingInterval + 50*time.Millisecond)
+	l := &loop{
+		pingInterval:  pingInterval,
+		pingTimer:     time.NewTimer(pingInterval),
+		intervalStart: now,
+	}
+	l.pingTimer.Stop()
+	done := l.pingTimerReset(false)
+	c.Assert(done, Equals, true)
+	select {
+	case <-l.pingTimer.C:
+	case <-time.After(1 * time.Second):
+		c.Fatal("timer should have triggered")
+	}
+	c.Check(now, Equals, l.intervalStart)
+}
+
+func (s *sessionSuite) TestPingTimerResetPartialTooLate(c *C) {
+	pingInterval := 5 * time.Second
+	now := time.Now().Add(-pingInterval - 50*time.Millisecond)
+	l := &loop{
+		pingInterval:  pingInterval,
+		pingTimer:     time.NewTimer(pingInterval),
+		intervalStart: now,
+	}
+	l.pingTimer.Stop()
+	done := l.pingTimerReset(false)
+	c.Assert(done, Equals, false)
+}
+
 func (s *sessionSuite) TestSessionLoopExchangeErrNop(c *C) {
 	nopTrack := NewTracker(s.testlog)
 	errCh := make(chan error, 1)
@@ -371,10 +426,50 @@ func (s *sessionSuite) TestSessionLoopExchangeErrNop(c *C) {
 	exchanges := make(chan broker.Exchange, 1)
 	exchanges <- &testExchange{prepErr: broker.ErrNop}
 	sess := &testing.TestBrokerSession{Exchanges: exchanges}
+	pingInterval := 5 * time.Second
 	go func() {
-		errCh <- sessionLoop(tp, sess, cfg5msPingInterval2msExchangeTout, nopTrack)
+		l := &loop{
+			proto:           tp,
+			sess:            sess,
+			track:           nopTrack,
+			pingInterval:    pingInterval,
+			pingTimer:       time.NewTimer(pingInterval),
+			exchangeTimeout: 1 * time.Second,
+			intervalStart:   time.Now().Add(-pingInterval + 50*time.Millisecond),
+		}
+		errCh <- l.run()
 	}()
-	c.Check(takeNext(down), Equals, "deadline 2ms")
+	c.Check(takeNext(down), Equals, "deadline 1s")
+	c.Check(takeNext(down), DeepEquals, protocol.PingPongMsg{Type: "ping"})
+	up <- nil // no write error
+	up <- io.EOF
+	err := <-errCh
+	c.Check(err, Equals, io.EOF)
+}
+
+func (s *sessionSuite) TestSessionLoopExchangeErrNopNeedPing(c *C) {
+	nopTrack := NewTracker(s.testlog)
+	errCh := make(chan error, 1)
+	up := make(chan interface{}, 5)
+	down := make(chan interface{}, 5)
+	tp := &testProtocol{up, down}
+	exchanges := make(chan broker.Exchange, 1)
+	exchanges <- &testExchange{prepErr: broker.ErrNop}
+	sess := &testing.TestBrokerSession{Exchanges: exchanges}
+	pingInterval := 5 * time.Second
+	go func() {
+		l := &loop{
+			proto:           tp,
+			sess:            sess,
+			track:           nopTrack,
+			pingInterval:    pingInterval,
+			pingTimer:       time.NewTimer(pingInterval),
+			exchangeTimeout: 1 * time.Second,
+			intervalStart:   time.Now().Add(-pingInterval - 50*time.Millisecond),
+		}
+		errCh <- l.run()
+	}()
+	c.Check(takeNext(down), Equals, "deadline 1s")
 	c.Check(takeNext(down), DeepEquals, protocol.PingPongMsg{Type: "ping"})
 	up <- nil // no write error
 	up <- io.EOF
@@ -477,18 +572,44 @@ func (s *sessionSuite) TestSessionLoopConnBrokenExchange(c *C) {
 	down := make(chan interface{}, 5)
 	tp := &testProtocol{up, down}
 	exchanges := make(chan broker.Exchange, 1)
-	exchanges <- &broker.ConnBrokenExchange{"REASON"}
+	msg := &protocol.ConnBrokenMsg{"connbroken", "BREASON"}
+	exchanges <- &broker.ConnMetaExchange{msg}
 	sess := &testing.TestBrokerSession{Exchanges: exchanges}
 	go func() {
 		errCh <- sessionLoop(tp, sess, cfg5msPingInterval2msExchangeTout, nopTrack)
 	}()
 	c.Check(takeNext(down), Equals, "deadline 2ms")
-	c.Check(takeNext(down), DeepEquals, protocol.ConnBrokenMsg{"connbroken", "REASON"})
+	c.Check(takeNext(down), DeepEquals, protocol.ConnBrokenMsg{"connbroken", "BREASON"})
 	up <- nil // no write error
 	err := <-errCh
 	c.Check(err, DeepEquals, &broker.ErrAbort{"session broken for reason"})
 }
 
+func (s *sessionSuite) TestSessionLoopConnWarnExchange(c *C) {
+	nopTrack := NewTracker(s.testlog)
+	errCh := make(chan error, 1)
+	up := make(chan interface{}, 5)
+	down := make(chan interface{}, 5)
+	tp := &testProtocol{up, down}
+	exchanges := make(chan broker.Exchange, 1)
+	msg := &protocol.ConnWarnMsg{"connwarn", "WREASON"}
+	exchanges <- &broker.ConnMetaExchange{msg}
+	sess := &testing.TestBrokerSession{Exchanges: exchanges}
+	go func() {
+		errCh <- sessionLoop(tp, sess, cfg5msPingInterval2msExchangeTout, nopTrack)
+	}()
+	c.Check(takeNext(down), Equals, "deadline 2ms")
+	c.Check(takeNext(down), DeepEquals, protocol.ConnWarnMsg{"connwarn", "WREASON"})
+	up <- nil // no write error
+	// session continues
+	c.Check(takeNext(down), Equals, "deadline 2ms")
+	c.Check(takeNext(down), DeepEquals, protocol.PingPongMsg{Type: "ping"})
+	up <- nil // no write error
+	up <- io.EOF
+	err := <-errCh
+	c.Check(err, Equals, io.EOF)
+}
+
 type testTracker struct {
 	SessionTracker
 	interval chan interface{}
@@ -526,7 +647,7 @@ func (s *sessionSuite) TestSessionLoopExchangeNextPing(c *C) {
 	effectiveOfPing := float64(interval) / float64(50*time.Millisecond)
 	comment := Commentf("effectiveOfPing=%f", effectiveOfPing)
 	c.Check(effectiveOfPing > 0.95, Equals, true, comment)
-	c.Check(effectiveOfPing < 1.15, Equals, true, comment)
+	c.Check(effectiveOfPing < 1.19, Equals, true, comment)
 	up <- nil // no write error
 	up <- io.EOF
 	err := <-errCh
@@ -593,7 +714,7 @@ func (s *sessionSuite) TestSessionWire(c *C) {
 	msg, err = downStream.ReadBytes(byte('}'))
 	c.Check(err, IsNil)
 	c.Check(msg, DeepEquals, []byte("\x00\x0c{\"T\":\"ping\"}"))
-	c.Check(takeNext(brkr.registration), Equals, "register DEV")
+	c.Check(takeNext(brkr.registration), Equals, "register DEV "+track.SessionId())
 	c.Check(len(brkr.registration), Equals, 0) // not yet unregistered
 	cli.Close()
 	err = <-errCh
